#include <pcl/point_types.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/cloud_viewer.h>
#include <pcl/console/print.h>
#include <pcl/console/print.h>

#include <chrono>
#include <ctime>

#include "entropy.h"
#include "binSegmentation.h"
#include "pose.h"
#include "wrinkle.h"

//----------------------------------------------------------------------------- //
int main(int argc, char **argv)
{

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr source(new pcl::PointCloud<pcl::PointXYZRGB>);
    if (pcl::io::loadPCDFile<pcl::PointXYZRGB>(argv[1], *source) == -1)
    {
        PCL_ERROR(" error opening file ");
        return (-1);
    }
    std::cout << "cloud orginal size: " << source->size() << std::endl;

    //time computation
    auto start = std::chrono::steady_clock::now();

    pcl::console::print_highlight("BinSegmentation...\n");
    //BIN SEGMENTATION -----------------------------------------------------------------------
    BinSegmentation bin;
    bin.setInputCloud(source);
    bin.setNumberLines(4);
    bin.setPaddingDistance(0.05); // 5cm from the bin walls
    bin.setMaxBinHeight(0.3);

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_grasp(new pcl::PointCloud<pcl::PointXYZRGB>);
    bool bin_result = bin.compute(cloud_grasp);
    if (bin_result == false)
        return -1;

    pcl::ModelCoefficients::Ptr plane = bin.getPlaneGroundPoints();
    pcl::PointCloud<pcl::PointXYZ>::Ptr top_vertices = bin.getVerticesBinContour();

    pcl::console::print_highlight("EntropyFilter...\n");
    // ENTROPY FILTER -----------------------------------------------------------------------
    //
    // ----------- ENTROPY FILTER -----------
    EntropyFilter ef;
    ef.setInputCloud(source);
    ef.setDownsampleLeafSize(0.005);
    ef.setEntropyThreshold(0.3);
    ef.setKLocalSearch(500);
    ef.setCurvatureThreshold(0.02);
    ef.setDepthThreshold(0.005);
    ef.setAngleThresholdForConvexity(10);
    ef.useWeightedEntropy(true);

    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> clouds_result;
    bool entropy_result = ef.compute(clouds_result);
    if (entropy_result == false)
        return -1;

    pcl::ModelCoefficients::Ptr plane_ef = ef.getReferencePlane();

    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_seg(new pcl::PointCloud<pcl::PointXYZI>);
    cloud_seg = ef.getCloudSeg();

    // ----------- WRINKLES DETECTOR -----------
    Poly py;
    py.setInputCloud(cloud_seg);
    py.setPlane(plane_ef);
    py.setRadius(0.02);
    py.setStep(0.02);
    py.setNumberIterations(100);

    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> wrinkles;
    py.compute(wrinkles);

    std::cout << "number of wrinkles found: " << wrinkles.size() << std::endl;

    // ----------- POSES ESTIMATION -----------
    Pose pose;
    pose.setSourceCloud(source);
    pose.setCloudSeg(cloud_seg);
    pose.setRefPlane(plane_ef);
    pose.setInputWrinklesVector(wrinkles);

    std::vector<Eigen::Affine3d, Eigen::aligned_allocator<Eigen::Affine3d>> matrices;
    pose.compute(matrices);

    ef.visualizeAll(false);

    return 0;
}
