#include <pcl/ModelCoefficients.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/filters/project_inliers.h>
#include <pcl/common/centroid.h>
#include <pcl/common/common.h>
#include <pcl/common/eigen.h>
#include <pcl/common/pca.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/segmentation/sac_segmentation.h>

#include <pcl/visualization/cloud_viewer.h>

typedef std::vector<Eigen::Affine3d, Eigen::aligned_allocator<Eigen::Affine3d>> Affine3dVector;
typedef std::vector<Eigen::Vector3f, Eigen::aligned_allocator<Eigen::Vector3f>> Vector3fVector;

struct CoordinateFrameT
{
    pcl::PointXYZ x;
    pcl::PointXYZ y;
    pcl::PointXYZ z;
    pcl::PointXYZ o;
};

class Pose
{

private:
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr m_source;
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> m_wrinkles, m_wrinkles_projected;
    pcl::PointCloud<pcl::PointXYZI>::Ptr m_cloud_seg;

    pcl::ModelCoefficients::Ptr m_plane;

public:
    Pose() : m_source(new pcl::PointCloud<pcl::PointXYZRGB>),
             m_plane(new pcl::ModelCoefficients),
             m_cloud_seg(new pcl::PointCloud<pcl::PointXYZI>)

    {
    }

    void setSourceCloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_in) { m_source = cloud_in; };

    void setCloudSeg(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_seg) { m_cloud_seg = cloud_seg; };

    void setRefPlane(pcl::ModelCoefficients::Ptr &plane) { m_plane = plane; };

    void setInputWrinklesVector(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &wrinkles) { m_wrinkles = wrinkles; };

    void compute(Affine3dVector &transformation_matrix_vector);

private:
    void projectWrinkle(pcl::PointCloud<pcl::PointXYZ>::Ptr &wrinkle_cloud,
                        pcl::PointCloud<pcl::PointXYZ>::Ptr &wrinkle_cloud_proj, pcl::ModelCoefficients::Ptr &plane);

    Eigen::Affine3d computeTransformation(Eigen::Vector3f &origin, Eigen::Vector3f &directionX, Eigen::Vector3f &directionZ);

    CoordinateFrameT createFrameForVisualization(Eigen::Vector3f &wrinkle_point, Eigen::Vector3f &wrinkle_dir, Eigen::Vector3f &normal_dir);

    bool ransacLineFitting(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::ModelCoefficients &line);

    void translatePoint(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_proj,
                        Eigen::Vector3f &point);

    void visualize(std::vector<CoordinateFrameT> &frames, std::vector<int> &skipped);
};