

#include <pcl/filters/project_inliers.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/extract_indices.h>

#include <pcl/common/centroid.h>
#include <pcl/common/common.h>
#include <pcl/common/eigen.h>
#include <pcl/common/pca.h>
#include <pcl/common/geometry.h>

#include <pcl/point_types.h>
#include <pcl/point_cloud.h>

#include <pcl/visualization/cloud_viewer.h>

struct PolypointsT
{
    pcl::PointXYZ search;
    pcl::PointXYZ centroid;
    pcl::PointXYZ pointComputed;
    pcl::PointXYZ pointParent;
};

class Poly
{

private:
    pcl::PointCloud<pcl::PointXYZI>::Ptr m_source, m_cloud;
    pcl::ModelCoefficients::Ptr m_plane;

    float m_radius, m_step;
    int m_iterations;

public:
    Poly() : m_source(new pcl::PointCloud<pcl::PointXYZI>),
             m_cloud(new pcl::PointCloud<pcl::PointXYZI>)
    {
    }

    void setNumberIterations(int N) { m_iterations = N; }

    void setRadius(float radius) { m_radius = radius; }

    void setStep(float step) { m_step = step; }

    void setInputCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud) { m_source = cloud; }

    void setPlane(pcl::ModelCoefficients::Ptr &plane) { m_plane = plane; }

    pcl::PointXYZ initialSearchPoint();

    pcl::PointXYZ newSearchPoint(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                                 pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_polyF,
                                 pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_polyB);

    pcl::PointCloud<pcl::PointXYZI>::Ptr computeRadiusNeighborhoodCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                                                                        pcl::PointXYZ &searchPoint, float radius);

    void computeDirection(pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_proj, Eigen::Vector3f &centroid, Eigen::Vector3f &direction);

    pcl::PointXYZ computeNewSearchPointBackward(Eigen::Vector3f &centroid, Eigen::Vector3f &direction, pcl::PointXYZ &parent, float step);

    pcl::PointXYZ computeNewSearchPointForward(Eigen::Vector3f &centroid, Eigen::Vector3f &direction, pcl::PointXYZ &parent, float step);

    bool checkSearchPoint(pcl::PointXYZ &searchPoint, float threshold);

    void computePointsInit(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud, PolypointsT &pointsF, PolypointsT &pointsB, float &radius, float &step);

    bool computeForward(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud, PolypointsT &points, float &radius, float &step);

    bool computeBackward(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud, PolypointsT &points, float &radius, float &step);

    void compute(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &output);

    void updatePointCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                          pcl::PointXYZ &point, float &radius);

    void visualization(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_F,
                       std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_B,
                       std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &output);

    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> prepareOutput(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_F,
                                                                   std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_B);
};